=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2006-02-21 22:41:44 +0000
+++ bzrlib/branch.py	2006-03-04 13:03:57 +0000
@@ -183,6 +183,14 @@
         """
         raise NotImplementedError('abspath is abstract')
 
+    def bind(self, other):
+        """Bind the local branch the other branch.
+
+        :param other: The branch to bind to
+        :type other: Branch
+        """
+        raise errors.UpgradeRequired(self.base)
+
     @needs_write_lock
     def fetch(self, from_branch, last_revision=None, pb=None):
         """Copy revisions from from_branch into this branch.
@@ -217,6 +225,21 @@
         finally:
             from_branch.unlock()
 
+    def get_bound_location(self):
+        """Return the URL of the rbanch we are bound to.
+
+        Older format branches cannot bind, please be sure to use a metadir
+        branch.
+        """
+        return None
+
+    def get_master_branch(self):
+        """Return the branch we are bound to.
+        
+        :return: Either a Branch, or None
+        """
+        return None
+
     def get_root_id(self):
         """Return the id of this branches root"""
         raise NotImplementedError('get_root_id is abstract')
@@ -243,6 +266,10 @@
         """
         return len(self.revision_history())
 
+    def unbind(self):
+        """Older format branches cannot bind or unbind."""
+        raise errors.UpgradeRequired(self.base)
+
     def last_revision(self):
         """Return last patch hash, or None if no history."""
         ph = self.revision_history()
@@ -251,7 +278,7 @@
         else:
             return None
 
-    def missing_revisions(self, other, stop_revision=None, diverged_ok=False):
+    def missing_revisions(self, other, stop_revision=None):
         """Return a list of new revisions that would perfectly fit.
         
         If self and other have not diverged, return a list of the revisions
@@ -301,7 +328,12 @@
         return other_history[self_len:stop_revision]
 
     def update_revisions(self, other, stop_revision=None):
-        """Pull in new perfect-fit revisions."""
+        """Pull in new perfect-fit revisions.
+
+        :param other: Another Branch to pull from
+        :param stop_revision: Updated until the given revision
+        :return: None
+        """
         raise NotImplementedError('update_revisions is abstract')
 
     def pullable_revisions(self, other, stop_revision):
@@ -656,10 +688,10 @@
             assert format.__class__ == self.__class__
         transport = a_bzrdir.get_branch_transport(None)
         control_files = LockableFiles(transport, 'lock')
-        return BzrBranch(_format=self,
-                         _control_files=control_files,
-                         a_bzrdir=a_bzrdir,
-                         _repository=a_bzrdir.find_repository())
+        return BzrBranch5(_format=self,
+                          _control_files=control_files,
+                          a_bzrdir=a_bzrdir,
+                          _repository=a_bzrdir.find_repository())
 
 
 class BranchReferenceFormat(BranchFormat):
@@ -744,17 +776,7 @@
     really matter if it's on an nfs/smb/afs/coda/... share, as long as
     it's writable, and can be accessed via the normal filesystem API.
     """
-    # We actually expect this class to be somewhat short-lived; part of its
-    # purpose is to try to isolate what bits of the branch logic are tied to
-    # filesystem access, so that in a later step, we can extricate them to
-    # a separarte ("storage") class.
-    _inventory_weave = None
     
-    # Map some sort of prefix into a namespace
-    # stuff like "revno:10", "revid:", etc.
-    # This should match a prefix with a function which accepts
-    REVISION_NAMESPACES = {}
-
     def __init__(self, transport=DEPRECATED_PARAMETER, init=DEPRECATED_PARAMETER,
                  relax_version_check=DEPRECATED_PARAMETER, _format=None,
                  _control_files=None, a_bzrdir=None, _repository=None):
@@ -901,7 +923,7 @@
         # TODO: test for failed two phase locks. This is known broken.
         self.repository.unlock()
         self.control_files.unlock()
-
+        
     def peek_lock_mode(self):
         if self.control_files._lock_count == 0:
             return None
@@ -979,7 +1001,6 @@
             self.append_revision(*pullable_revs)
 
     def pullable_revisions(self, other, stop_revision):
-        """See Branch.pullable_revisions."""
         other_revno = other.revision_id_to_revno(stop_revision)
         try:
             return self.missing_revisions(other, other_revno)
@@ -1063,39 +1084,133 @@
     def tree_config(self):
         return TreeConfig(self)
 
-    def _get_truncated_history(self, revision_id):
-        history = self.revision_history()
-        if revision_id is None:
-            return history
+
+class BzrBranch5(BzrBranch):
+    """A format 5 branch. This supports new features over plan branches.
+
+    It has support for a master_branch which is the data for bound branches.
+    """
+
+    def __init__(self,
+                 _format,
+                 _control_files,
+                 a_bzrdir,
+                 _repository):
+        super(BzrBranch5, self).__init__(_format=_format,
+                                         _control_files=_control_files,
+                                         a_bzrdir=a_bzrdir,
+                                         _repository=_repository)
+        
+    @needs_write_lock
+    def pull(self, source, overwrite=False, stop_revision=None):
+        """Updates branch.pull to be bound branch aware."""
+        # TODO: jam 20051230 This does work, in that 'bzr pull'
+        #       will update the master branch before updating the
+        #       local branch. However, 'source' can also already
+        #       be the master branch. Which means that we are
+        #       asking it to update from itself, before we continue.
+        #       This probably causes double downloads, etc.
+        #       So we probably want to put in an explicit check
+        #       of whether source is already the master branch.
+        bound_location = self.get_bound_location()
+        if source.base != bound_location:
+            # not pulling from master, so we need to update master.
+            master_branch = self.get_master_branch()
+            if master_branch:
+                master_branch.pull(source)
+                source = master_branch
+        return super(BzrBranch5, self).pull(source, overwrite, stop_revision)
+
+    def get_bound_location(self):
         try:
-            idx = history.index(revision_id)
-        except ValueError:
-            raise InvalidRevisionId(revision_id=revision, branch=self)
-        return history[:idx+1]
+            return self.control_files.get_utf8('bound').read()[:-1]
+        except errors.NoSuchFile:
+            return None
 
     @needs_read_lock
-    def _clone_weave(self, to_location, revision=None, basis_branch=None):
-        # prevent leakage
-        from bzrlib.workingtree import WorkingTree
-        assert isinstance(to_location, basestring)
-        if basis_branch is not None:
-            note("basis_branch is not supported for fast weave copy yet.")
-
-        history = self._get_truncated_history(revision)
-        if not bzrlib.osutils.lexists(to_location):
-            os.mkdir(to_location)
-        bzrdir_to = self.bzrdir._format.initialize(to_location)
-        self.repository.clone(bzrdir_to)
-        branch_to = bzrdir_to.create_branch()
-        mutter("copy branch from %s to %s", self, branch_to)
-
-        # FIXME duplicate code with base .clone().
-        # .. would template method be useful here?  RBC 20051207
-        branch_to.set_parent(self.base)
-        branch_to.append_revision(*history)
-        WorkingTree.create(branch_to, branch_to.base)
-        mutter("copied")
-        return branch_to
+    def get_master_branch(self):
+        """Return the branch we are bound to.
+        
+        :return: Either a Branch, or None
+
+        This could memoise the branch, but if thats done
+        it must be revalidated on each new lock.
+        So for now we just dont memoise it.
+        # RBC 20060304 review this decision.
+        """
+        bound_loc = self.get_bound_location()
+        if not bound_loc:
+            return None
+        try:
+            return Branch.open(bound_loc)
+        except (errors.NotBranchError, errors.ConnectionError), e:
+            raise errors.BoundBranchConnectionFailure(
+                    self, bound_loc, e)
+
+    @needs_write_lock
+    def set_bound_location(self, location):
+        """Set the target where this branch is bound to.
+
+        :param location: URL to the target branch
+        """
+        if location:
+            self.control_files.put_utf8('bound', location+'\n')
+        else:
+            try:
+                self.control_files._transport.delete('bound')
+            except NoSuchFile:
+                return False
+            return True
+
+    @needs_write_lock
+    def bind(self, other):
+        """Bind the local branch the other branch.
+
+        :param other: The branch to bind to
+        :type other: Branch
+        """
+        # TODO: jam 20051230 Consider checking if the target is bound
+        #       It is debatable whether you should be able to bind to
+        #       a branch which is itself bound.
+        #       Committing is obviously forbidden,
+        #       but binding itself may not be.
+        #       Since we *have* to check at commit time, we don't
+        #       *need* to check here
+        self.pull(other)
+
+        # we are now equal to or a suffix of other.
+
+        # Since we have 'pulled' from the remote location,
+        # now we should try to pull in the opposite direction
+        # in case the local tree has more revisions than the
+        # remote one.
+        # There may be a different check you could do here
+        # rather than actually trying to install revisions remotely.
+        # TODO: capture an exception which indicates the remote branch
+        #       is not writeable. 
+        #       If it is up-to-date, this probably should not be a failure
+        
+        # lock other for write so the revision-history syncing cannot race
+        other.lock_write()
+        try:
+            other.pull(self)
+            # if this does not error, other now has the same last rev we do
+            # it can only error if the pull from other was concurrent with
+            # a commit to other from someone else.
+
+            # until we ditch revision-history, we need to sync them up:
+            self.set_revision_history(other.revision_history())
+            # now other and self are up to date with each other and have the
+            # same revision-history.
+        finally:
+            other.unlock()
+
+        self.set_bound_location(other.base)
+
+    @needs_write_lock
+    def unbind(self):
+        """If bound, unbind"""
+        return self.set_bound_location(None)
 
 
 class BranchTestProviderAdapter(object):

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-03-03 05:38:04 +0000
+++ bzrlib/builtins.py	2006-03-04 13:03:57 +0000
@@ -549,10 +549,11 @@
     branch before copying anything from the remote branch.
     """
     takes_args = ['from_location', 'to_location?']
-    takes_options = ['revision', 'basis']
+    takes_options = ['revision', 'basis', 'bound']
     aliases = ['get', 'clone']
 
-    def run(self, from_location, to_location=None, revision=None, basis=None):
+    def run(self, from_location, to_location=None, revision=None, basis=None,
+            bound=False):
         if revision is None:
             revision = [None]
         elif len(revision) > 1:
@@ -604,7 +605,7 @@
                 raise BzrCommandError(msg)
             except bzrlib.errors.UnlistableBranch:
                 rmtree(to_location)
-                msg = "The branch %s cannot be used as a --basis"
+                msg = "The branch %s cannot be used as a --basis" % (basis,)
                 raise BzrCommandError(msg)
             if name:
                 branch.control_files.put_utf8('branch-name', name)
@@ -612,6 +613,8 @@
             note('Branched %d revision(s).' % branch.revno())
         finally:
             br_from.unlock()
+        if bound:
+            branch.bind(br_from)
 
 
 class cmd_checkout(Command):
@@ -1436,6 +1439,10 @@
         except StrictCommitFailed:
             raise BzrCommandError("Commit refused because there are unknown "
                                   "files in the working tree.")
+        except errors.BoundBranchOutOfDate, e:
+            raise BzrCommandError(str(e)
+                                  + ' Either unbind or update.')
+
         note('Committed revision %d.' % (tree.branch.revno(),))
 
 
@@ -2122,6 +2129,68 @@
                 raise BzrCommandError('Please supply either one revision, or a range.')
 
 
+class cmd_bind(Command):
+    """Bind the current branch to its parent.
+
+    After binding, commits must succeed on the parent branch
+    before they can be done on the local one.
+    """
+
+    takes_args = ['location?']
+    takes_options = []
+
+    def run(self, location=None):
+        b, relpath = Branch.open_containing(u'.')
+        if location is None:
+            location = b.get_bound_location()
+        if location is None:
+            location = b.get_parent()
+        if location is None:
+            raise BzrCommandError('Branch has no parent,'
+                                  ' you must supply a bind location.')
+        b_other = Branch.open(location)
+        try:
+            b.bind(b_other)
+        except DivergedBranches:
+            raise BzrCommandError('These branches have diverged.'
+                                  ' Try merging, and then bind again.')
+
+
+class cmd_unbind(Command):
+    """Bind the current branch to its parent.
+
+    After unbinding, the local branch is considered independent.
+    """
+
+    takes_args = []
+    takes_options = []
+
+    def run(self):
+        b, relpath = Branch.open_containing(u'.')
+        if not b.unbind():
+            raise BzrCommandError('Local branch is not bound')
+
+
+class cmd_update(Command):
+    """Update the local tree for checkouts and bound branches.
+    """
+    def run(self):
+        wt, relpath = WorkingTree.open_containing(u'.')
+        # TODO: jam 20051127 Check here to see if this is a checkout
+        bound_loc = wt.branch.get_bound_location()
+        if not bound_loc:
+            raise BzrCommandError('Working tree %s is not a checkout'
+                                  ' or a bound branch, you probably'
+                                  ' want pull' % wt.base)
+
+        br_bound = Branch.open(bound_loc)
+        try:
+            wt.pull(br_bound, overwrite=False)
+        except DivergedBranches:
+            raise BzrCommandError("These branches have diverged."
+                                  "  Try merge.")
+
+
 class cmd_uncommit(bzrlib.commands.Command):
     """Remove the last committed revision.
 

=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2006-02-23 04:08:56 +0000
+++ bzrlib/commit.py	2006-03-04 13:03:57 +0000
@@ -224,6 +224,8 @@
             raise BzrError("The message keyword parameter is required for commit().")
 
         self.weave_store = self.branch.repository.weave_store
+        self.bound_branch = None
+        self.master_branch = None
         self.rev_id = rev_id
         self.specific_files = specific_files
         self.allow_pointless = allow_pointless
@@ -231,47 +233,52 @@
         if revprops:
             self.revprops.update(revprops)
 
-        # check for out of date working trees
-        if self.work_tree.last_revision() != self.branch.last_revision():
-            raise errors.OutOfDateTree(self.work_tree)
-
-        if strict:
-            # raise an exception as soon as we find a single unknown.
-            for unknown in self.work_tree.unknowns():
-                raise StrictCommitFailed()
-
-        if timestamp is None:
-            self.timestamp = time.time()
-        else:
-            self.timestamp = long(timestamp)
-            
-        if self.config is None:
-            self.config = bzrlib.config.BranchConfig(self.branch)
-
-        if rev_id is None:
-            self.rev_id = _gen_revision_id(self.config, self.timestamp)
-        else:
-            self.rev_id = rev_id
-
-        if committer is None:
-            self.committer = self.config.username()
-        else:
-            assert isinstance(committer, basestring), type(committer)
-            self.committer = committer
-
-        if timezone is None:
-            self.timezone = local_time_offset()
-        else:
-            self.timezone = int(timezone)
-
-        if isinstance(message, str):
-            message = message.decode(bzrlib.user_encoding)
-        assert isinstance(message, unicode), type(message)
-        self.message = message
-        self._escape_commit_message()
-
-        self.branch.lock_write()
+        self.work_tree.lock_write()
         try:
+            # setup the bound branch variables as needed.
+            self._check_bound_branch()
+
+            # check for out of date working trees
+            # if we are bound, then self.branch is the master branch and this
+            # test is thus all we need.
+            if self.work_tree.last_revision() != self.branch.last_revision():
+                raise errors.OutOfDateTree(self.work_tree)
+    
+            if strict:
+                # raise an exception as soon as we find a single unknown.
+                for unknown in self.work_tree.unknowns():
+                    raise StrictCommitFailed()
+    
+            if timestamp is None:
+                self.timestamp = time.time()
+            else:
+                self.timestamp = long(timestamp)
+                
+            if self.config is None:
+                self.config = bzrlib.config.BranchConfig(self.branch)
+    
+            if rev_id is None:
+                self.rev_id = _gen_revision_id(self.config, self.timestamp)
+            else:
+                self.rev_id = rev_id
+    
+            if committer is None:
+                self.committer = self.config.username()
+            else:
+                assert isinstance(committer, basestring), type(committer)
+                self.committer = committer
+    
+            if timezone is None:
+                self.timezone = local_time_offset()
+            else:
+                self.timezone = int(timezone)
+    
+            if isinstance(message, str):
+                message = message.decode(bzrlib.user_encoding)
+            assert isinstance(message, unicode), type(message)
+            self.message = message
+            self._escape_commit_message()
+
             self.work_inv = self.work_tree.inventory
             self.basis_tree = self.work_tree.basis_tree()
             self.basis_inv = self.basis_tree.inventory
@@ -300,13 +307,22 @@
                 self.present_parents
                 )
             self._make_revision()
+            # revision is in the master branch now.
+            
+            self.branch.append_revision(self.rev_id)
+            # now its in the master branch history.
+
+            self._update_bound_branch()
+            # now the local branch is up to date
+
             self.work_tree.set_pending_merges([])
-            self.branch.append_revision(self.rev_id)
             if len(self.parents):
                 precursor = self.parents[0]
             else:
                 precursor = None
             self.work_tree.set_last_revision(self.rev_id, precursor)
+            # now the work tree is up to date with the branch
+            
             self.reporter.completed(self.branch.revno()+1, self.rev_id)
             if self.config.post_commit() is not None:
                 hooks = self.config.post_commit().split(' ')
@@ -317,8 +333,83 @@
                                    'bzrlib':bzrlib,
                                    'rev_id':self.rev_id})
         finally:
-            self.branch.unlock()
-
+            self._cleanup_bound_branch()
+            self.work_tree.unlock()
+
+    def _check_bound_branch(self):
+        """Check to see if the local branch is bound.
+
+        If it is bound, then most of the commit will actually be
+        done using the remote branch as the target branch.
+        Only at the end will the local branch be updated.
+        """
+        # TODO: jam 20051230 Consider a special error for the case
+        #       where the local branch is bound, and can't access the
+        #       master branch
+        self.master_branch = self.branch.get_master_branch()
+        if not self.master_branch:
+            return
+
+        # If the master branch is bound, we must fail
+        master_bound_location = self.master_branch.get_bound_location()
+        if master_bound_location:
+            raise errors.CommitToDoubleBoundBranch(self.branch,
+                    self.master_branch, master_bound_location)
+
+        # TODO: jam 20051230 We could automatically push local
+        #       commits to the remote branch if they would fit.
+        #       But for now, just require remote to be identical
+        #       to local.
+        
+        # Make sure the local branch is identical to the master
+        master_rh = self.master_branch.revision_history()
+        local_rh = self.branch.revision_history()
+        if local_rh != master_rh:
+            raise errors.BoundBranchOutOfDate(self.branch,
+                    self.master_branch)
+
+        # Now things are ready to change the master branch
+        # so grab the lock
+        self.bound_branch = self.branch
+        self.master_branch.lock_write()
+        self.branch = self.master_branch
+        
+        # Check to see if we have any pending merges. If we do
+        # those need to be pushed into the master branch
+        pending_merges = self.work_tree.pending_merges()
+        if pending_merges:
+            for revision_id in pending_merges:
+                self.master_branch.repository.fetch(self.bound_branch.repository,
+                                                    revision_id=revision_id)
+
+    def _cleanup_bound_branch(self):
+        """Executed at the end of a try/finally to cleanup a bound branch.
+
+        If the branch wasn't bound, this is a no-op.
+        If it was, it resents self.branch to the local branch, instead
+        of being the master.
+        """
+        if not self.bound_branch:
+            return
+        self.branch = self.bound_branch
+        self.master_branch.unlock()
+
+    def _update_bound_branch(self):
+        """Update the local bound branch, after commit.
+
+        This only runs if the commit to the master branch succeeds.
+        """
+        if not self.bound_branch:
+            return
+        # We always want the local branch to look like the remote one
+        # TODO: jam 20051231 We might want overwrite=True here, but
+        #       the local branch should be a prefix of master anyway
+        self.bound_branch.pull(self.master_branch)
+
+        # TODO: jam 20051231 At this point we probably 
+        #       want to merge any changes into master branch's
+        #       working tree.
+        
     def _escape_commit_message(self):
         """Replace xml-incompatible control characters."""
         # Python strings can include characters that can't be

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-02-26 07:54:02 +0000
+++ bzrlib/errors.py	2006-03-04 13:03:57 +0000
@@ -368,9 +368,11 @@
         self.format = format
 
 
+
 class StrictCommitFailed(Exception):
     """Commit refused because there are unknowns in the tree."""
 
+
 class NoSuchRevision(BzrError):
     def __init__(self, branch, revision):
         self.branch = branch
@@ -415,6 +417,7 @@
         BzrError.__init__(self, msg)
 
 
+
 class NotAncestor(BzrError):
     def __init__(self, rev_id, not_ancestor_id):
         msg = "Revision %s is not an ancestor of %s" % (not_ancestor_id, 
@@ -450,11 +453,45 @@
         BzrError.__init__(self, "Store %s is not listable" % store)
 
 
+
 class UnlistableBranch(BzrError):
     def __init__(self, br):
         BzrError.__init__(self, "Stores for branch %s are not listable" % br)
 
 
+class BoundBranchOutOfDate(BzrNewError):
+    """Bound branch %(branch)s is out of date with master branch %(master)s."""
+    def __init__(self, branch, master):
+        BzrNewError.__init__(self)
+        self.branch = branch
+        self.master = master
+
+        
+class CommitToDoubleBoundBranch(BzrNewError):
+    """Cannot commit to branch %(branch)s. It is bound to %(master)s, which is bound to %(remote)s."""
+    def __init__(self, branch, master, remote):
+        BzrNewError.__init__(self)
+        self.branch = branch
+        self.master = master
+        self.remote = remote
+
+
+class OverwriteBoundBranch(BzrNewError):
+    """Cannot pull --overwrite to a branch which is bound %(branch)s"""
+    def __init__(self, branch):
+        BzrNewError.__init__(self)
+        self.branch = branch
+
+
+class BoundBranchConnectionFailure(BzrNewError):
+    """Unable to connect to target of bound branch %(branch)s => %(target)s: %(error)s"""
+    def __init__(self, branch, target, error):
+        BzrNewError.__init__(self)
+        self.branch = branch
+        self.target = target
+        self.error = error
+
+
 class WeaveError(BzrNewError):
     """Error in processing weave: %(message)s"""
     def __init__(self, message=None):
@@ -618,9 +655,11 @@
         self.text_revision = text_revision
         self.file_id = file_id
 
+
 class DuplicateKey(BzrNewError):
     """Key %(key)s is already present in map"""
 
+
 class MalformedTransform(BzrNewError):
     """Tree transform is malformed %(conflicts)r"""
 
@@ -709,6 +748,7 @@
         BzrNewError.__init__(self)
         self.tree = tree
 
+
 class CorruptRepository(BzrNewError):
     """An error has been detected in the repository %(repo_path)s.
 Please run bzr reconcile on this repository."""
@@ -716,3 +756,12 @@
     def __init__(self, repo):
         BzrNewError.__init__(self)
         self.repo_path = repo.bzrdir.root_transport.base
+
+
+class UpgradeRequired(BzrNewError):
+    """To use this feature you must upgrade your branch at %(path)s."""
+
+    def __init__(self, path):
+        BzrNewError.__init__(self)
+        self.path = path
+

=== modified file 'bzrlib/option.py'
--- bzrlib/option.py	2006-02-03 19:50:59 +0000
+++ bzrlib/option.py	2006-03-04 13:03:57 +0000
@@ -161,6 +161,7 @@
 _global_option('overwrite', help='Ignore differences between branches and '
                'overwrite unconditionally')
 _global_option('basis', type=str)
+_global_option('bound')
 _global_option('diff-options', type=str)
 _global_option('help',
                help='show help message')
@@ -179,6 +180,7 @@
 _global_option('timezone', 
                type=str,
                help='display timezone as local, original, or utc')
+_global_option('unbound')
 _global_option('verbose',
                help='display more information')
 _global_option('version')

=== modified file 'bzrlib/tests/blackbox/__init__.py'
--- bzrlib/tests/blackbox/__init__.py	2006-03-03 05:55:51 +0000
+++ bzrlib/tests/blackbox/__init__.py	2006-03-04 13:03:57 +0000
@@ -37,6 +37,7 @@
                      'bzrlib.tests.blackbox.test_added',
                      'bzrlib.tests.blackbox.test_aliases',
                      'bzrlib.tests.blackbox.test_ancestry',
+                     'bzrlib.tests.blackbox.test_bound_branches',
                      'bzrlib.tests.blackbox.test_cat',
                      'bzrlib.tests.blackbox.test_checkout',
                      'bzrlib.tests.blackbox.test_commit',

=== added file 'bzrlib/tests/blackbox/test_bound_branches.py'
--- bzrlib/tests/blackbox/test_bound_branches.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/blackbox/test_bound_branches.py	2006-03-04 13:03:57 +0000
@@ -0,0 +1,388 @@
+# Copyright (C) 2005 by Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+
+"""Tests of bound branches (binding, unbinding, commit, etc) command.
+"""
+
+import os
+from cStringIO import StringIO
+
+from bzrlib.tests import TestCaseWithTransport
+from bzrlib.branch import Branch
+from bzrlib.bzrdir import (BzrDir,
+                           BzrDirFormat,
+                           BzrDirFormat6,
+                           BzrDirMetaFormat1,
+                           )
+from bzrlib.workingtree import WorkingTree
+
+
+class TestLegacyFormats(TestCaseWithTransport):
+    
+    def setUp(self):
+        super(TestLegacyFormats, self).setUp()
+        self.build_tree(['master/', 'child/'])
+        self.run_bzr('init', 'master')
+        old_format = BzrDirFormat.get_default_format()
+        BzrDirFormat.set_default_format(BzrDirFormat6())
+        try:
+            self.run_bzr('init', 'child')
+        finally:
+            BzrDirFormat.set_default_format(old_format)
+        os.chdir('child')
+    
+    def test_bind_format_6_bzrdir(self):
+        # bind on a format 6 bzrdir should error
+        out,err = self.run_bzr('bind', '../master', retcode=3)
+        self.assertEqual('', out)
+        self.assertEqual('bzr: ERROR: To use this feature you must '
+                         'upgrade your branch at %s/.\n' % os.getcwdu(), err)
+    
+    def test_unbind_format_6_bzrdir(self):
+        # bind on a format 6 bzrdir should error
+        out,err = self.run_bzr('unbind', retcode=3)
+        self.assertEqual('', out)
+        self.assertEqual('bzr: ERROR: To use this feature you must '
+                         'upgrade your branch at %s/.\n' % os.getcwdu(), err)
+
+
+class TestBoundBranches(TestCaseWithTransport):
+
+    def create_branches(self):
+        bzr = self.run_bzr
+        self.build_tree(['base/', 'base/a', 'base/b'])
+
+        self.init_meta_branch('base')
+        os.chdir('base')
+        bzr('add')
+        bzr('commit', '-m', 'init')
+
+        os.chdir('..')
+
+        bzr('get', '--bound', 'base', 'child')
+
+        self.failUnlessExists('child')
+
+        self.check_revno(1, 'child')
+        d = BzrDir.open('child')
+        self.assertNotEqual(None, d.open_branch().get_master_branch())
+
+    def check_revno(self, val, loc=None):
+        if loc is not None:
+            cwd = os.getcwd()
+            os.chdir(loc)
+        self.assertEquals(str(val), self.run_bzr('revno')[0].strip())
+        if loc is not None:
+            os.chdir(cwd)
+
+    def test_simple_binding(self):
+        self.build_tree(['base/', 'base/a', 'base/b'])
+
+        self.init_meta_branch('base')
+        self.run_bzr('add', 'base')
+        self.run_bzr('commit', '-m', 'init', 'base')
+
+        self.run_bzr('branch', 'base', 'child')
+
+        os.chdir('child')
+        self.run_bzr('bind', '../base')
+
+        d = BzrDir.open('')
+        self.assertNotEqual(None, d.open_branch().get_master_branch())
+
+        self.run_bzr('unbind')
+        self.assertEqual(None, d.open_branch().get_master_branch())
+
+        self.run_bzr('unbind', retcode=3)
+
+    def init_meta_branch(self, path):
+        old_format = BzrDirFormat.get_default_format()
+        BzrDirFormat.set_default_format(BzrDirMetaFormat1())
+        try:
+            self.run_bzr('init', path)
+        finally:
+            BzrDirFormat.set_default_format(old_format)
+
+    def test_bound_commit(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('child')
+        open('a', 'wb').write('new contents\n')
+        bzr('commit', '-m', 'child')
+
+        self.check_revno(2)
+
+        # Make sure it committed on the parent
+        self.check_revno(2, '../base')
+
+    def test_bound_fail(self):
+        # Make sure commit fails if out of date.
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('base')
+        open('a', 'wb').write('new base contents\n')
+        bzr('commit', '-m', 'base')
+
+        os.chdir('../child')
+        open('b', 'wb').write('new b child contents\n')
+        bzr('commit', '-m', 'child', retcode=3)
+
+        bzr('update')
+        self.check_revno(2)
+
+        bzr('commit', '-m', 'child')
+        self.check_revno(3)
+        self.check_revno(3, '../base')
+
+    def test_double_binding(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        bzr('branch', 'child', 'child2')
+        os.chdir('child2')
+
+        # Double binding succeeds, but committing to child2 should fail
+        bzr('bind', '../child')
+
+        bzr('commit', '-m', 'child2', '--unchanged', retcode=3)
+
+    def test_unbinding(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('base')
+        open('a', 'wb').write('new base contents\n')
+        bzr('commit', '-m', 'base')
+        self.check_revno(2)
+
+        os.chdir('../child')
+        open('b', 'wb').write('new b child contents\n')
+        self.check_revno(1)
+        bzr('commit', '-m', 'child', retcode=3)
+        self.check_revno(1)
+        bzr('unbind')
+        bzr('commit', '-m', 'child')
+        self.check_revno(2)
+
+        bzr('bind', retcode=3)
+
+    def test_commit_remote_bound(self):
+        # It is not possible to commit to a branch
+        # which is bound to a branch which is bound
+        bzr = self.run_bzr
+        self.create_branches()
+        bzr('branch', 'base', 'newbase')
+        os.chdir('base')
+        
+        # There is no way to know that B has already
+        # been bound by someone else, otherwise it
+        # might be nice if this would fail
+        bzr('bind', '../newbase')
+
+        os.chdir('../child')
+        bzr('commit', '-m', 'failure', '--unchanged', retcode=3)
+
+    def test_pull_updates_both(self):
+        bzr = self.run_bzr
+        self.create_branches()
+        bzr('branch', 'base', 'newchild')
+        os.chdir('newchild')
+        open('b', 'wb').write('newchild b contents\n')
+        bzr('commit', '-m', 'newchild')
+        self.check_revno(2)
+
+        os.chdir('../child')
+        # The pull should succeed, and update
+        # the bound parent branch
+        bzr('pull', '../newchild')
+        self.check_revno(2)
+
+        self.check_revno(2, '../base')
+
+    def test_bind_diverged(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('child')
+        bzr('unbind')
+
+        bzr('commit', '-m', 'child', '--unchanged')
+        self.check_revno(2)
+
+        os.chdir('../base')
+        self.check_revno(1)
+        bzr('commit', '-m', 'base', '--unchanged')
+        self.check_revno(2)
+
+        os.chdir('../child')
+        # These branches have diverged
+        bzr('bind', '../base', retcode=3)
+
+        # TODO: In the future, this might require actual changes
+        # to have occurred, rather than just a new revision entry
+        bzr('merge', '../base')
+        bzr('commit', '-m', 'merged')
+        self.check_revno(3)
+
+        # After a merge, trying to bind again should succeed
+        # by pushing the new change to base
+        bzr('bind', '../base')
+        self.check_revno(3)
+        self.check_revno(3, '../base')
+
+        # After binding, the revision history should be identical
+        child_rh = bzr('revision-history')[0]
+        os.chdir('../base')
+        base_rh = bzr('revision-history')[0]
+        self.assertEquals(child_rh, base_rh)
+
+    def test_bind_parent_ahead(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('child')
+        bzr('unbind')
+
+        os.chdir('../base')
+        bzr('commit', '-m', 'base', '--unchanged')
+
+        os.chdir('../child')
+        self.check_revno(1)
+        bzr('bind', '../base')
+
+        self.check_revno(2)
+        bzr('unbind')
+
+        # Check and make sure it also works if parent is ahead multiple
+        os.chdir('../base')
+        bzr('commit', '-m', 'base 3', '--unchanged')
+        bzr('commit', '-m', 'base 4', '--unchanged')
+        bzr('commit', '-m', 'base 5', '--unchanged')
+        self.check_revno(5)
+
+        os.chdir('../child')
+        self.check_revno(2)
+        bzr('bind')
+        self.check_revno(5)
+
+    def test_bind_child_ahead(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        os.chdir('child')
+        bzr('unbind')
+        bzr('commit', '-m', 'child', '--unchanged')
+        self.check_revno(2)
+        self.check_revno(1, '../base')
+
+        bzr('bind', '../base')
+        self.check_revno(2, '../base')
+
+        # Check and make sure it also works if child is ahead multiple
+        bzr('unbind')
+        bzr('commit', '-m', 'child 3', '--unchanged')
+        bzr('commit', '-m', 'child 4', '--unchanged')
+        bzr('commit', '-m', 'child 5', '--unchanged')
+        self.check_revno(5)
+
+        self.check_revno(2, '../base')
+        bzr('bind')
+        self.check_revno(5, '../base')
+
+    def test_commit_after_merge(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        # We want merge to be able to be a local only
+        # operation, because it can be without violating
+        # the binding invariants.
+        # But we can't fail afterwards
+
+        bzr('branch', 'child', 'other')
+
+        os.chdir('other')
+        open('c', 'wb').write('file c\n')
+        bzr('add', 'c')
+        bzr('commit', '-m', 'adding c')
+        new_rev_id = bzr('revision-history')[0].strip().split('\n')[-1]
+
+        os.chdir('../child')
+        bzr('merge', '../other')
+
+        self.failUnlessExists('c')
+        tree = WorkingTree.open('.')
+        self.assertEqual([new_rev_id], tree.pending_merges())
+
+        # Make sure the local branch has the installed revision
+        bzr('cat-revision', new_rev_id)
+        
+        # And make sure that the base tree does not
+        os.chdir('../base')
+        bzr('cat-revision', new_rev_id, retcode=3)
+
+        # Commit should succeed, and cause merged revisions to
+        # be pulled into base
+        os.chdir('../child')
+        bzr('commit', '-m', 'merge other')
+
+        self.check_revno(2)
+
+        os.chdir('../base')
+        self.check_revno(2)
+
+        bzr('cat-revision', new_rev_id)
+
+    def test_pull_overwrite_fails(self):
+        bzr = self.run_bzr
+        self.create_branches()
+
+        bzr('branch', 'child', 'other')
+        
+        os.chdir('other')
+        open('a', 'wb').write('new contents\n')
+        bzr('commit', '-m', 'changed a')
+        self.check_revno(2)
+        open('a', 'ab').write('and then some\n')
+        bzr('commit', '-m', 'another a')
+        self.check_revno(3)
+        open('a', 'ab').write('and some more\n')
+        bzr('commit', '-m', 'yet another a')
+        self.check_revno(4)
+
+        os.chdir('../child')
+        open('a', 'wb').write('also changed a\n')
+        bzr('commit', '-m', 'child modified a')
+
+        self.check_revno(2)
+        self.check_revno(2, '../base')
+
+        # It might be possible that we want pull --overwrite to
+        # actually succeed.
+        # If we want it, just change this test to make sure that 
+        # both base and child are updated properly
+        bzr('pull', '--overwrite', '../other', retcode=3)
+
+        # It should fail without changing the local revision
+        self.check_revno(2)
+        self.check_revno(2, '../base')
+
+    # TODO: jam 20051230 Test that commit & pull fail when the branch we 
+    #       are bound to is not available
+
+

=== modified file 'bzrlib/tests/branch_implementations/__init__.py'
--- bzrlib/tests/branch_implementations/__init__.py	2006-02-10 11:03:38 +0000
+++ bzrlib/tests/branch_implementations/__init__.py	2006-03-04 13:03:57 +0000
@@ -39,6 +39,7 @@
 def test_suite():
     result = TestSuite()
     test_branch_implementations = [
+        'bzrlib.tests.branch_implementations.test_bound_sftp',
         'bzrlib.tests.branch_implementations.test_branch',
         'bzrlib.tests.branch_implementations.test_parent',
         'bzrlib.tests.branch_implementations.test_permissions',

=== added file 'bzrlib/tests/branch_implementations/test_bound_sftp.py'
--- bzrlib/tests/branch_implementations/test_bound_sftp.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/branch_implementations/test_bound_sftp.py	2006-03-04 13:03:57 +0000
@@ -0,0 +1,366 @@
+# Copyright (C) 2005 Robey Pointer <robey@lag.net>, Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for branches bound to an sftp branch."""
+
+
+import os
+
+
+from bzrlib.branch import Branch
+from bzrlib.bzrdir import (BzrDir,
+                           BzrDirFormat,
+                           BzrDirFormat6,
+                           BzrDirMetaFormat1,
+                           )
+import bzrlib.errors as errors
+from bzrlib.tests.test_sftp_transport import TestCaseWithSFTPServer, paramiko_loaded
+
+
+class BoundSFTPBranch(TestCaseWithSFTPServer):
+
+    def create_branches(self):
+        self.build_tree(['base/', 'base/a', 'base/b'])
+        old_format = BzrDirFormat.get_default_format()
+        BzrDirFormat.set_default_format(BzrDirMetaFormat1())
+        try:
+            wt_base = BzrDir.create_standalone_workingtree('base')
+        finally:
+            BzrDirFormat.set_default_format(old_format)
+    
+        b_base = wt_base.branch
+
+        wt_base.add('a')
+        wt_base.add('b')
+        wt_base.commit('first', rev_id='r@b-1')
+
+        wt_child = b_base.bzrdir.sprout('child').open_workingtree()
+        self.sftp_base = Branch.open(self.get_url('base'))
+        wt_child.branch.bind(self.sftp_base)
+
+        self.assertEqual(['r@b-1'], b_base.revision_history())
+        self.assertEqual(['r@b-1'], wt_child.branch.revision_history())
+
+        return b_base, wt_child
+
+    def test_simple_binding(self):
+        self.build_tree(['base/', 'base/a', 'base/b', 'child/'])
+        wt_base = BzrDir.create_standalone_workingtree('base')
+
+        wt_base.add('a')
+        wt_base.add('b')
+        wt_base.commit('first', rev_id='r@b-1')
+
+        b_base = wt_base.branch
+        old_format = BzrDirFormat.get_default_format()
+        BzrDirFormat.set_default_format(BzrDirMetaFormat1())
+        try:
+            b_child = BzrDir.create_branch_convenience('child')
+        finally:
+            BzrDirFormat.set_default_format(old_format)
+        self.assertEqual(None, b_child.get_bound_location())
+        self.assertEqual(None, b_child.get_master_branch())
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+        b_child.bind(sftp_b_base)
+        self.assertEqual(sftp_b_base.base, b_child.get_bound_location())
+        # this line is more of a working tree test line, but - what the hey.
+        b_child.bzrdir.open_workingtree().update()
+        self.failUnlessExists('child/a')
+        self.failUnlessExists('child/b')
+
+        b_child.unbind()
+        self.assertEqual(None, b_child.get_bound_location())
+
+    def test_bound_commit(self):
+        b_base, wt_child = self.create_branches()
+
+        open('child/a', 'wb').write('new contents\n')
+        wt_child.commit('modified a', rev_id='r@c-2')
+
+        self.assertEqual(['r@b-1', 'r@c-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@c-2'], b_base.revision_history())
+
+    def test_bound_fail(self):
+        # Make sure commit fails if out of date.
+        b_base, wt_child = self.create_branches()
+
+        open('base/a', 'wb').write('new base contents\n')
+        b_base.bzrdir.open_workingtree().commit('base', rev_id='r@b-2')
+
+        open('child/b', 'wb').write('new b child contents\n')
+        self.assertRaises(errors.BoundBranchOutOfDate,
+                wt_child.commit, 'child', rev_id='r@c-2')
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+
+        # This is all that cmd_update does
+        wt_child.pull(sftp_b_base, overwrite=False)
+
+        wt_child.commit('child', rev_id='r@c-3')
+
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@c-3'],
+                wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@c-3'],
+                b_base.revision_history())
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@c-3'],
+                sftp_b_base.revision_history())
+
+    def test_double_binding(self):
+        b_base, wt_child = self.create_branches()
+
+        wt_child2 = wt_child.bzrdir.sprout('child2').open_workingtree()
+
+        wt_child2.branch.bind(wt_child.branch)
+
+        open('child2/a', 'wb').write('new contents\n')
+        self.assertRaises(errors.CommitToDoubleBoundBranch,
+                wt_child2.commit, 'child2', rev_id='r@d-2')
+
+    def test_unbinding(self):
+        from bzrlib.transport import get_transport
+        b_base, wt_child = self.create_branches()
+
+        # TestCaseWithSFTPServer only allows you to connect one time
+        # to the SFTP server. So we have to create a connection and
+        # keep it around, so that it can be reused
+        __unused_t = get_transport(self.get_url('.'))
+
+        wt_base = b_base.bzrdir.open_workingtree()
+        open('base/a', 'wb').write('new base contents\n')
+        wt_base.commit('base', rev_id='r@b-2')
+
+        open('child/b', 'wb').write('new b child contents\n')
+        self.assertRaises(errors.BoundBranchOutOfDate,
+                wt_child.commit, 'child', rev_id='r@c-2')
+        self.assertEqual(['r@b-1'], wt_child.branch.revision_history())
+        wt_child.branch.unbind()
+        wt_child.commit('child', rev_id='r@c-2')
+        self.assertEqual(['r@b-1', 'r@c-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@b-2'], b_base.revision_history())
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+        self.assertRaises(errors.DivergedBranches,
+                wt_child.branch.bind, sftp_b_base)
+
+    def test_commit_remote_bound(self):
+        # Make sure it is detected if the current base
+        # suddenly is bound when child goes to commit
+        b_base, wt_child = self.create_branches()
+
+        b_base.bzrdir.sprout('newbase')
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+        sftp_b_newbase = Branch.open(self.get_url('newbase'))
+
+        sftp_b_base.bind(sftp_b_newbase)
+
+        open('child/a', 'wb').write('new contents\n')
+        self.assertRaises(errors.CommitToDoubleBoundBranch,
+                wt_child.commit, 'failure', rev_id='r@c-2')
+
+        self.assertEqual(['r@b-1'], b_base.revision_history())
+        self.assertEqual(['r@b-1'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1'], sftp_b_newbase.revision_history())
+
+    def test_pull_updates_both(self):
+        b_base, wt_child = self.create_branches()
+
+        wt_newchild = b_base.bzrdir.sprout('newchild').open_workingtree()
+        open('newchild/b', 'wb').write('newchild b contents\n')
+        wt_newchild.commit('newchild', rev_id='r@d-2')
+        self.assertEqual(['r@b-1', 'r@d-2'], wt_newchild.branch.revision_history())
+
+        wt_child.pull(wt_newchild.branch)
+        self.assertEqual(['r@b-1', 'r@d-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@d-2'], b_base.revision_history())
+
+    def test_bind_diverged(self):
+        from bzrlib.builtins import merge
+
+        b_base, wt_child = self.create_branches()
+
+        wt_child.branch.unbind()
+        open('child/a', 'ab').write('child contents\n')
+        wt_child.commit('child', rev_id='r@c-2')
+
+        self.assertEqual(['r@b-1', 'r@c-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1'], b_base.revision_history())
+
+        open('base/b', 'ab').write('base contents\n')
+        b_base.bzrdir.open_workingtree().commit('base', rev_id='r@b-2')
+        self.assertEqual(['r@b-1', 'r@b-2'], b_base.revision_history())
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+
+        self.assertRaises(errors.DivergedBranches,
+                wt_child.branch.bind, sftp_b_base)
+
+        # TODO: jam 20051230 merge_inner doesn't set pending merges
+        #       Is this on purpose?
+        #       merge_inner also doesn't fetch any missing revisions
+        #merge_inner(wt_child.branch, sftp_b_base.revision_tree('r@b-2'), 
+        #        wt_child.branch.revision_tree('r@b-1'))
+        # TODO: jam 20051230 merge(..., (None, None), ...) seems to
+        #       cause an infinite loop of some sort. It definitely doesn't
+        #       work, you have to use list notation
+        merge((sftp_b_base.base, 2), [None, None], this_dir=wt_child.branch.base)
+
+        self.assertEqual(['r@b-2'], wt_child.pending_merges())
+        wt_child.commit('merged', rev_id='r@c-3')
+
+        # After a merge, trying to bind again should succeed
+        # by pushing the new change to base
+        wt_child.branch.bind(sftp_b_base)
+
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@c-3'],
+                b_base.revision_history())
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@c-3'],
+                wt_child.branch.revision_history())
+
+    def test_bind_parent_ahead(self):
+        b_base, wt_child = self.create_branches()
+
+        wt_child.branch.unbind()
+
+        open('a', 'ab').write('base changes\n')
+        wt_base = b_base.bzrdir.open_workingtree()
+        wt_base.commit('base', rev_id='r@b-2')
+        self.assertEqual(['r@b-1', 'r@b-2'], b_base.revision_history())
+        self.assertEqual(['r@b-1'], wt_child.branch.revision_history())
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+        wt_child.branch.bind(sftp_b_base)
+
+        self.assertEqual(['r@b-1', 'r@b-2'], wt_child.branch.revision_history())
+
+        wt_child.branch.unbind()
+
+        # Check and make sure it also works if parent is ahead multiple
+        wt_base.commit('base 3', rev_id='r@b-3', allow_pointless=True)
+        wt_base.commit('base 4', rev_id='r@b-4', allow_pointless=True)
+        wt_base.commit('base 5', rev_id='r@b-5', allow_pointless=True)
+
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@b-3', 'r@b-4', 'r@b-5'],
+                b_base.revision_history())
+
+        self.assertEqual(['r@b-1', 'r@b-2'], wt_child.branch.revision_history())
+
+        wt_child.branch.bind(sftp_b_base)
+        self.assertEqual(['r@b-1', 'r@b-2', 'r@b-3', 'r@b-4', 'r@b-5'],
+                wt_child.branch.revision_history())
+
+    def test_bind_child_ahead(self):
+        b_base, wt_child = self.create_branches()
+
+        wt_child.branch.unbind()
+
+        wt_child.commit('child', rev_id='r@c-2', allow_pointless=True)
+        self.assertEqual(['r@b-1', 'r@c-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1'], b_base.revision_history())
+
+        sftp_b_base = Branch.open(self.get_url('base'))
+        wt_child.branch.bind(sftp_b_base)
+
+        self.assertEqual(['r@b-1', 'r@c-2'], b_base.revision_history())
+
+        # Check and make sure it also works if child is ahead multiple
+        wt_child.branch.unbind()
+        wt_child.commit('child 3', rev_id='r@c-3', allow_pointless=True)
+        wt_child.commit('child 4', rev_id='r@c-4', allow_pointless=True)
+        wt_child.commit('child 5', rev_id='r@c-5', allow_pointless=True)
+
+        self.assertEqual(['r@b-1', 'r@c-2', 'r@c-3', 'r@c-4', 'r@c-5'],
+                wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@c-2'], b_base.revision_history())
+
+        wt_child.branch.bind(sftp_b_base)
+        self.assertEqual(['r@b-1', 'r@c-2', 'r@c-3', 'r@c-4', 'r@c-5'],
+                b_base.revision_history())
+
+    def test_commit_after_merge(self):
+        from bzrlib.builtins import merge
+
+        b_base, wt_child = self.create_branches()
+
+        # We want merge to be able to be a local only
+        # operation, because it does not alter the branch data.
+
+        # But we can't fail afterwards
+
+        wt_other = wt_child.bzrdir.sprout('other').open_workingtree()
+
+        open('other/c', 'wb').write('file c\n')
+        wt_other.add('c')
+        wt_other.commit('adding c', rev_id='r@d-2')
+
+        self.failIf(wt_child.branch.repository.has_revision('r@d-2'))
+        self.failIf(b_base.repository.has_revision('r@d-2'))
+
+        # TODO: jam 20051230 merge_inner doesn't set pending merges
+        #       Is this on purpose?
+        #       merge_inner also doesn't fetch any missing revisions
+        #merge_inner(wt_child.branch, b_other.revision_tree('r@d-2'),
+        #        wt_child.branch.revision_tree('r@b-1'))
+        merge((wt_other.branch.base, 2), [None, None], this_dir=wt_child.branch.base)
+
+        self.failUnlessExists('child/c')
+        self.assertEqual(['r@d-2'], wt_child.pending_merges())
+        self.failUnless(wt_child.branch.repository.has_revision('r@d-2'))
+        self.failIf(b_base.repository.has_revision('r@d-2'))
+
+        # Commit should succeed, and cause merged revisions to
+        # be pulled into base
+        wt_child.commit('merge other', rev_id='r@c-2')
+        self.assertEqual(['r@b-1', 'r@c-2'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@c-2'], b_base.revision_history())
+        self.failUnless(b_base.repository.has_revision('r@d-2'))
+
+    def test_commit_fails(self):
+        b_base, wt_child = self.create_branches()
+
+        open('a', 'ab').write('child adds some text\n')
+
+        del b_base
+        os.rename('base', 'hidden_base')
+
+        self.assertRaises(errors.BoundBranchConnectionFailure,
+                wt_child.commit, 'added text', rev_id='r@c-2')
+
+    def test_pull_fails(self):
+        b_base, wt_child = self.create_branches()
+
+        wt_other = wt_child.bzrdir.sprout('other').open_workingtree()
+        open('other/a', 'wb').write('new contents\n')
+        wt_other.commit('changed a', rev_id='r@d-2')
+
+        self.assertEqual(['r@b-1'], b_base.revision_history())
+        self.assertEqual(['r@b-1'], wt_child.branch.revision_history())
+        self.assertEqual(['r@b-1', 'r@d-2'], wt_other.branch.revision_history())
+
+        del b_base
+        os.rename('base', 'hidden_base')
+
+        self.assertRaises(errors.BoundBranchConnectionFailure,
+                wt_child.pull, wt_other.branch)
+
+    # TODO: jam 20051231 We need invasive failure tests, so that we can show
+    #       performance even when something fails.
+
+
+if not paramiko_loaded:
+    del BoundSFTPBranch
+

=== modified file 'bzrlib/tests/test_source.py'
--- bzrlib/tests/test_source.py	2006-02-11 11:58:06 +0000
+++ bzrlib/tests/test_source.py	2006-03-04 13:03:57 +0000
@@ -72,4 +72,4 @@
         # written. Note that this is an exact equality so that when the number
         # drops, it is not given a buffer but rather this test updated
         # immediately.
-        self.assertEqual(6, occurences)
+        self.assertEqual(7, occurences)

